"""
	IQsubmit: A program to submit batches of tasks to a SLURM controller
    Copyright (C) 2017  Alexandre Foley 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

import shutil
import os
import configparser
import typing
import re
import collections

sbatch_opt = 'sbatch options'
larray = "--array"
sarray = '-a'

Origin_fold = os.getcwd()

class IQError(Exception):
    def __init__(self, message):

	        # Call the base class constructor with the parameters it needs
			        super(Exception, self).__init__(message)


def WARNING(*stringin):
	print('\x1b[0;31;40m', *stringin , '\x1b[0m')

def parsefile(str_filename):
	config = configparser.ConfigParser(allow_no_value=True,interpolation=None)
	config.read(str_filename)
	terminate = False
	if sbatch_opt not in config:
		print("No [sbatch options] section.")
		terminate = True
	else:
		if len(config[sbatch_opt])==0:
			terminate = True
			print("No options for sbatch. At least --job-name must be specified")
		else :
			if not "--job-name" in config[sbatch_opt]:
				print('No specified jobname')
				terminate = True
			if not "--array" in config[sbatch_opt]  and "-a" in config[sbatch_opt]:
				#Internally only --array is used. both are otherwise valid.
				config[sbatch_opt]['--array'] = config[sbatch_opt]['-a']
				config.remove_option(sbatch_opt, '-a')
			if not ('-o' in config[sbatch_opt] or '--output' in config[sbatch_opt]):
				#write default output in the right folder for each job by default.
				config.set(sbatch_opt,'--output','job%a.QJ/slurm-%A_%a.out')
			if '--wrap' in config[sbatch_opt]:
				terminate=True
				print ('Do not use the sbatch option "--wrap" with IQsubmit. \nUse the command section instead.')
	if 'command' not in config:
		print ( "No [command] section.")
		terminate = True
	else:
		if len(config['command'])==0:
			print("No command to execute.")
			terminate = True
		if 'copy_loop' in config['command']:
			opt_val = config['command'].get('copyfiles','') 
			if opt_val !='':
				opt_val += ','
			opt_val += config[sbatch_opt]['--job-name']+'/'+config['command'].get('copy_loop_name','cloop.dat') 
			config('command','copyfiles',opt_val)
	if 'loop' not in config:
			print("No [loop] section.")
			#terminate = True
	else:
		if len(config['loop'])==0:
			print("Nothing in [loop] section. A loop must be specified")
			#terminate = True
		elif not all(x in [ 'loop', 'loop_file'] for x in config['loop']):
			print('Unknown option in loop section. Known options are loop and loop_file')
			terminate = True

	if terminate:
		WARNING( "Invalid qcmbatch file, terminating immediatly.")
	return [terminate,config]


#Flatten a list of list (but leaves strings alone)
def flatten(l):
	for el in l:
		if isinstance(el, collections.Iterable) and not isinstance(el, (str, bytes)):
			yield from flatten(el)
		else:
			yield el


def arrayopt2index(string_array: str) -> typing.List[str]:
	#We throw away the maximum number of job at once specifier
	string_array = string_array.split('%')[0]

	string_array = string_array.split(',')
	for i,el in enumerate(string_array):
		if '-' in el:
			#split the string on '-' or ':'
			tmp = re.split('-|:',el)
			if len(tmp) == 3:
				string_array[i] = [str(r) for r in range(int(tmp[0]),int(tmp[1])+1,int(tmp[2]))]
			else:
				string_array[i] = [str(r) for r in range(int(tmp[0]),int(tmp[1])+1)]
				
	return flatten(string_array)


def recur_loop(desc):
	the_rest = desc[1:]
	steps_val = [i[1] for i in desc[0][0:-1]]
	out = [i[0] for i in desc[0][0:-1]]
	for i in range(desc[0][-1]):
		if len(desc)>1:
			for line in recur_loop(desc[1:]):
				yield out + line
		else:
			yield out
		out = [val+step_val for val,step_val in zip(out,steps_val ) ]
			

def parse_desc(desc):
	desc = desc.split(',')
	header = []
	loop_len = 1
	for j,subdesc in enumerate(desc):
		subdesc, step = subdesc.split('):')
		step = int(step)
		loop_len *= step
		subdesc = subdesc.split(')')
		for i,el in enumerate(subdesc):
			tmp = re.split('\(|:',el)
			header += [tmp[0]]
			subdesc[i] = [float(tmp[1]),(float(tmp[2])-float(tmp[1]))/(step-1)]
		subdesc.append(step)
		desc[j] = subdesc
	return header,desc,loop_len
	
def loop_from_desc(desc):
	header,pdesc,llen = parse_desc(desc)
	yield header
	for line in recur_loop(pdesc):
		yield line


def gen_loop(loop_section):
	header = ""
	item = []
	do_loop  = False
	if "loop_file" in loop_section:
		do_loop = True
		full_loop = open(loop_section["loop_file"],'r')
		header = full_loop.readline().strip()
		for line in full_loop:
			item.append(line.strip().split('\t'))
	elif "loop" in loop_section:
		do_loop = True
		header_l,descs,len_loop = parse_desc(loop_section["loop"])
		header = ''
		for name in header_l:
			header+= name + '\t'
		for line in recur_loop(descs):
			item.append(line)
	return do_loop,header,item

def string_sub(old:str,new:str,string:str) -> str:
	return string.replace(old,new)

def linspace(start, stop, n):
	if n == 1:
		yield stop
		return
	h = (stop - start) / (n - 1)
	for i in range(n):
		yield start + h * i

def parse_substitution(config,indexes):
	list_sub = 0
	list_header	= 0
	substitutions = 0
	substitution_fn = 0
	do_substitute = False
	do_file_sub = False
	do_list_sub = False
	do_sub = False
	header=''
	data = []
	if 'simple' == config['command'].get('substitution method','simple'):
		substitution_fn = string_sub
	elif 'regex' ==  config['command'].get('substitution method'):
		import re.sub as sub
		substitution_fn = sub
	else:
		raise IQError("Unknown replacement procedure. Supported values for substitution are 'substitution method = regex' and 'substitution method = simple'.")
	if 'substitute' in config['command']:
		do_substitute = True
		do_sub = True
		substitutions = config['command']['substitute'].split(',')
		for i,item in enumerate(substitutions):
			substitutions[i] = item.split('=')
			substitutions[i][0] = substitutions[i][0].rstrip().lstrip()
			substitutions[i][1] = re.sub(r'%u',r'{username}',substitutions[i][1])
			substitutions[i][1] = re.sub(r'%x',r'{jobname}',substitutions[i][1])
			substitutions[i][1] = re.sub(r'%(\d*)a',r'{ArrayInd:0=\1}',substitutions[i][1])
	if 'file_substitute' in config['command']:
		do_substitute = True
		do_file_sub = True
		if 'list_substitute' in config['command']:
			pass
#			raise RuntimeError("Both list_substitute and file substitute are present in [command] section. Only one at a time is suppported.")
		sub_file = open(os.path.join(Origin_fold,config['command']['file_substitute']),'r')
		header = sub_file.readline().split()
		header = [h.rstrip().lstrip() for h in header]
		data = [line.split() for line in sub_file]
		if len(data) != len(indexes):
			if len(data) < len(indexes):
				print(len(indexes))
				print(len(data))
				raise IQError("More task than substitution entry in " + config['command']['file_substitute'] + '!')
			WARNING('WARNING: more substition entry than tasks to execute, not all the substitutions in '+ config['command']['file_substitute']+ ' are used.')
	if 'list_substitute' in config['command']:
		do_substitute = True
		do_list_sub = True
		#print( config['command']['list_substitute'])
		#list_sub = config['command']['list_substitute'].rstrip(')').split(')')
		#print(list_sub)
		list_header, descs, len_loop = parse_desc(config['command']['list_substitute'].strip())
		if len_loop != len(indexes):
			if len_loop < len(indexes):
				raise IQError("More task ("+str(len(indexes)) +") than substitution entry in list_substitute = " + config['command']['list_substitute'] + '!')
			WARNING('WARNING: more substition entry than tasks to execute, not all the substitutions in '+ config['command']['list_substitute']+ ' are used.')
		list_sub = [line for line in recur_loop(descs)]
	return substitution_fn,do_substitute,do_sub,do_file_sub,do_list_sub,substitutions,header,data,list_header,list_sub


def template_sub(config,indexes,parsed_bundle):
	[substitution_fn,do_substitute,do_sub,do_file_sub,do_list_sub,substitutions,header,data,list_header,list_sub] = parsed_bundle
	if do_substitute:
		for i,ind in enumerate(indexes):
			for template_file in [x for x in config['command']['template'].split(',') if x]:
				template_file = open(os.path.join('job'+ind+'.QJ', config['command']['template']),'r').read()
				if do_sub:
					for sub in substitutions:
						frt_str = sub[1].format(username=os.getenv('USER'),jobname=config[sbatch_opt]['--job-name'],ArrayInd=int(ind))
						#print('sub: ',sub[0],' ', frt_str)
						template_file = substitution_fn('~'+sub[0]+'~',frt_str,template_file)
				if do_file_sub:
					for old_str,new_str in zip(header,data[i]):
						#print('file_sub: ', old_str,' ' ,new_str)
						template_file = substitution_fn('~'+old_str+'~',new_str,template_file)
				if do_list_sub:
					for (old_str,list_new) in zip(list_header,list_sub[i]):
						#print('list_sub: ', old_str,' ' ,list_new[i])
						template_file = substitution_fn('~'+old_str+'~',str(list_new),template_file)
				open(os.path.join('job'+ind+'.QJ', os.path.split(config['command']['template'])[-1]),'w').write(template_file)
			
def compute_size_from_substitute(config):
	fn = ln = 'undefined'
	if 'file_substitute' in config['command']:
		fn = 0 
		for line in open(config['command']['file_substitute'],'r'):
			fn+=1
	if 'list_substitute' in config['command']:
		ln = 1
		A = config['command']['list_substitute'].split(',')
		for ite in A:
			ln*= int(ite.split(':')[-1])
		if fn == 'undefined':
			return ln-1
	if ln == 'undefined':
		return fn-2
	else:
		return min(ln-1,fn-2)

def check_files_exists(config,or_fol):
	def actual_check(path,short_name):
		if not (os.path.isfile(path) or os.path.isdir(path)):
			raise IQError("The file or directory "+ short_name + " does not exist in the current directory.")

	if 'linkfiles' in config['command']:
		for linksrc in config['command'].get('linkfiles').split(','):
			actual_check(os.path.join(or_fol,linksrc),linksrc)
	if 'template' in config['command']:
		for tpltsrc in config['command'].get('template').split(','):
			actual_check(os.path.join(or_fol,tpltsrc),tpltsrc)
	if 'copyfiles' in config['command']:
		for copysrc in config['command'].get('copyfiles').split(','):
			actual_check(os.path.join(or_fol,copysrc),copysrc)


def generate_files(config):
	#change to the specified workdir if there's one.
	#Throw and error if the folder doesn't exists.
	os.chdir(config[sbatch_opt].get('-D',config[sbatch_opt].get('--workdir','.')))
	if config[sbatch_opt].get('--array','auto')=='auto':
		indexes = [x for x in arrayopt2index('0-'+str(compute_size_from_substitute(config)))]
		config[sbatch_opt]['--array'] = str(indexes[0])+'-'+str(indexes[-1])
	else:
		indexes = [x for x in arrayopt2index(config[sbatch_opt].get('--array','0'))]
	if 'template' in config['command']:
		#ALL parsing that might lead to failure must be done before creating any file or directory.
		#That way, the failure doesn't leave stuff in a partially done state.
		subs_parse_bundle = parse_substitution(config,indexes)
	if "loop" in config:
		do_loop,loop_header,loop_items = gen_loop(config['loop'])
		if len(loop_items)<len(indexes):
			raise IQError("The loop to split in [loop] has less element than there are task!")
	check_files_exists(config,Origin_fold)
	try:
		os.makedirs(config[sbatch_opt]['--job-name'])
	except FileExistsError:
		WARNING('directory "'+config[sbatch_opt]['--job-name'].strip()+'" already exist.')
	os.chdir(config[sbatch_opt]['--job-name'])
	if 'link_loop' in config['command']:
		llop = 	open(config['command'].get('link_loop_name','lloop.dat'),'w')
		for line in loop_from_desc(config['command']['link_loop']):
			for word in line:
				llop.write(str(word))
				llop.write('\t')
			llop.write('\n')
		llop.close()
		opt_val = config['command'].get('linkfiles','') 
		if opt_val !='':
			opt_val += ','
		opt_val += config[sbatch_opt]['--job-name']+'/'+config['command'].get('link_loop_name','lloop.dat') 
		config.set('command','linkfiles',opt_val)
	if 'copy_loop' in config['command']:
		clop = 	open(config['command'].get('copy_loop_name','lloop.dat'),'w')
		for line in loop_from_desc(config['command']['copy_loop']):
			for word in line:
				clop.write(str(word))
				clop.write('\t')
			clop.write('\n')
		clop.close()
	loop_step = len(loop_items)//len(indexes)
	loop_mod = len(loop_items)%len(indexes)
	dist = 0
	items_start = 0
	rel_fol = os.path.relpath(Origin_fold,os.getcwd())
	for i,ind in enumerate(indexes):
		dirname = "job"+ind+".QJ"
		os.makedirs(dirname)
		if 'linkfiles' in config['command']:
			for linksrc in config['command'].get('linkfiles').split(','):
				os.symlink(os.path.join(rel_fol,'..',linksrc),os.path.join(dirname,os.path.split(linksrc)[-1]))
		if 'template' in config['command']:
			for tpltsrc in config['command'].get('template').split(','):
				shutil.copy(os.path.join(Origin_fold,tpltsrc),os.path.join(dirname,os.path.split(tpltsrc)[-1]))
		if 'copyfiles' in config['command']:
			for copysrc in config['command'].get('copyfiles').split(','):
				shutil.copy(os.path.join(Origin_fold,copysrc),os.path.join(dirname,os.path.split(copysrc)[-1]))
		#if need arise, other ways to substitute the values for the loop will be designed.
		if do_loop:
			loop_file = open(os.path.join(dirname,'loop.dat'),'w')
			loop_file.write(loop_header+'\n')
			if i < loop_mod:
			#make sure the remainder is evenly distributed
				dist+=1
			items_end = min(loop_step*(i+1)+dist,len(loop_items))
			for item_line in loop_items[items_start:items_end]:
				loop_file.write('\t'.join([str(x) for x in item_line])+'\n')
			items_start = items_end
	if 'template' in config['command']:
		template_sub(config,indexes,subs_parse_bundle)

def prepare_export_env(config):
	#set the number of OpenMP to cpus-per-task threads for if it's not set explicitly
	#set it explicitly if you want a number of OMP thread different from the number of cpus-per-task asked for
	#I think this is only reasonable for whole node computations.
	export_str = config[sbatch_opt].get('--export','')
	omp_thread = config[sbatch_opt].getint('--cpus-per-task',config[sbatch_opt].getint('-c',1))
	if not 'OMP_NUM_THREADS' in export_str:
		export_str += 'OMP_NUM_THREADS='+str(omp_thread)
	config.set(sbatch_opt,'--export',export_str) 

def gen_command(command):
	command_str = command['command']
	##no you cannot submit a job to erase stuff!
	assert(not 'rm' is command_str.lstrip()[:2])
	command_str += ' ' + command.get('options','')
	return command_str

def sbatch_opt_string(config):
	out_str = ''
	for key in config[sbatch_opt]:
		out_str += key
		if config[sbatch_opt][key] !=None:
			spacer = ' ' if (len(key) == 2)  else '='
			out_str += spacer + config[sbatch_opt][key]
		out_str += ' '
	out_str += '--wrap=\'cd job${SLURM_ARRAY_TASK_ID}.QJ; '+gen_command(config['command'])+';\''
	return out_str

def call_sbatch(config_file_name,dryrun = False):
	terminate, config = parsefile(config_file_name)
	if terminate:
		print ('terminated')
		return
	generate_files(config)
	prepare_export_env(config)
	sbatch_opt_str = sbatch_opt_string(config)
	if not dryrun: 
		os.system('sbatch '+sbatch_opt_str)
	else:
		print('sbatch ' + sbatch_opt_str)


if __name__=='__main__':
	call_sbatch('qcmsub',True)



